--=======================================================================
-- miniSIPServer IVR-XML script
--
-- Copyright (C) MYVOIPAPP,Inc. All Rights Reserved.
--
-- author: Gilson
--
-- History:
-- 2016-08-17 migrated from Python script
--=======================================================================

require "services"
require "utils.xml"
require "utils.dba"

-- MSS inner IVR-XML services.
-- key should be same with Exml_services defined in mss core.
local MSS_INNER_IVRXML={[1]="aa.xml"}

-- IVR action methods
local IVR_METHOD_PNC      = "playandwaitinput"
local IVR_METHOD_PA       = "playannouncement"
local IVR_METHOD_CALLTO   = "callto"
local IVR_METHOD_INPUT    = "input"
local IVR_METHOD_RELCALL  = "releasecall"
local IVR_METHOD_GETURL   = "geturl"
local IVR_METHOD_GETURL_RESULT = "geturlresult"

-- FSM service state
local EIVR_IDLE = 0
local EIVR_INTERACTION = 1
local EIVR_WAIT_ANSWER = 2

-- default timer (seconds)
local IVR_TIMER_INTERACTION = 120
local IVR_TIMER_WAIT_ANSWER = 120

-- MAX value
local MAX_FAIL_CONN_CNT = 10

--====================================
--
-- database functions
--
--====================================
local function dbGetCustIVRXML(dialPlan, calledNbr)
    local sqlStr = string.format("select fileName from tbl_ivr_trigger where"..
                   "  dialPlan='%s' and  calledNbr='%s'", dialPlan, calledNbr)                  
    local result = cdbQuery(sqlStr)
    if result~=MSS_ERR_OBJ_FOUND then
        return -1
    end

    fsm.ivrxmlFileName = cdbGetRowStrVal(0)
    return 1
end

--====================================
--
-- FSM common function
--
--====================================
local function initData()
    fsm.dialPlan = fsm.origDialPlan
    fsm.ivrInd = MSS_NO
    fsm.pncResult = ""
    fsm.srfError = ESRF_ERR_OK
    fsm.ivrxmlFileName = ""
    fsm.geturlResult = ""
    fsm.connFailCnt = 0 -- count of failed connect operation

    fsm.previousActionIndex = 0
    fsm.previousActionMethod = ""
    fsm.currActionIndex = 1
    fsm.currActionMethod = ""
    fsm.actionCnt=0
    fsm.actionDict = {}  -- use action 'name' as key, store parsed actions
    fsm.actionNameList = {}  --  use index to search action name
    fsm.xmlDoc=0 -- XML document handler
end


local function exception()
    SendReleaseCall()
    EndService()
end


local function getCustIVRXML()
    local ivrID = currMessage.ivrXMLServiceID
    if ivrID~=MSS_INVALID_OBJ_ID then
        fsm.ivrxmlFileName = MSS_INNER_IVRXML[ivrID]
    end

    local result = dbGetCustIVRXML(fsm.dialPlan, fsm.calledNbr)
    if result<0 and fsm.dialPlan~=MSS_DEFAULT_DIAL_PLAN then
        dbGetCustIVRXML(MSS_DEFAULT_DIAL_PLAN, fsm.calledNbr)
    end

    if string.len(fsm.ivrxmlFileName)==0 then
        TraceInfo("\t cannot find IVR XML file.")
        return -1
    end

    TraceInfo("\t current IVR-XML is "..fsm.ivrxmlFileName)
    return 1
end 


local function monitorCall()
    TraceInfo("\t monitorCall")
    SendRRBE(CALLED_LEGID, DP_O_CALL_PROGRESS, DP_MON_INTERUPT)
    SendRRBE(CALLED_LEGID, DP_O_ANSWER, DP_MON_INTERUPT)
end


local function prepareIVR()
    TraceInfo("\t prepareIVR")
    if fsm.ivrInd==MSS_NO then
        SendCTR(CALLER_LEGID)
        fsm.ivrInd = MSS_YES
    end
end


local function stopIVR()
    TraceInfo("\t stopIVR")
    if fsm.ivrInd==MSS_YES then
        SendDFC(CALLER_LEGID)
        fsm.ivrInd = MSS_NO
    end
end


local function getInputActionResult(actionName)
    local inputAction = fsm.actionDict[actionName]
    if inputAction.method~=IVR_METHOD_INPUT then
        return ""
    else
        if inputAction.result~=nil then
            return inputAction.result
        else
            return ""
        end
    end
end


local function getInputResult(inputNode)
    if inputNode.from~=nil then
        return getInputActionResult(inputNode.from)
    else
        return fsm.pncResult -- always get current PnC result
    end    
end


local function getFullURL(currAction) -- currAction should be 'geturl'
    local fullURL = currAction.url
    local paramCnt = currAction.paramCnt
    if paramCnt==nil or paramCnt<=0 then
        TraceInfo("\t\t fullURL: "..fullURL)
        return fullURL
    end
    
    local paramStrs="?"
    for index=1, paramCnt do
        local param = currAction.params[index]
        local valueName=param.name        
        local value = param.value
        local valueType = value.type
        local valueContent = ""
        if valueType=="stringVal" then
            valueContent = value.content
        elseif valueType=="callernbr" then
            valueContent = fsm.callerNbr            
        elseif valueType=="callednbr" then
            valueContent = fsm.calledNbr
        elseif valueType=="inputVal" then
            valueContent = getInputResult(value.content)
        else
            TraceInfo("\t\t unknown parameter type: "..valueType)
            return ""
        end

        paramStrs = paramStrs..valueName.."="..valueContent
        if index<paramCnt then
            paramStrs = paramStrs.."&"
        end
    end

    fullURL=fullURL..paramStrs
    TraceInfo("\t\t fullURL: "..fullURL)
    return fullURL
end

--====================================
--
-- Parser functions
--
--====================================
local function annIDToInt(annID)
    -- annID must be a string
    if annID==nil or string.len(annID)==0 then return 0 end
    local annStr = string.lower(annID)
    local result = string.find(annStr, "x")
    if result==nil then
        annStr = "0x"..annStr
    end
    return ToInt(annStr)
end


local function parserElementAudio()
    TraceInfo("\t\t parserElementAudio")
    local ivrAudio={}
    local eErrNo = xmlChildElementByName(fsm.xmlDoc,"id")
    if eErrNo~=MSS_ERR_OK then 
        TraceInfo("\t\t\t cannot find 'id' element")
        return nil 
    end    
    ivrAudio.id = annIDToInt(xmlGetElementData(fsm.xmlDoc))

    -- back to audio element to search 'duration' element
    xmlParentElement(fsm.xmlDoc)
    eErrNo = xmlChildElementByName(fsm.xmlDoc,"duration")
    if eErrNo==MSS_ERR_OK then
        ivrAudio.duration = ToInt(xmlGetElementData(fsm.xmlDoc))
        xmlParentElement(fsm.xmlDoc) -- back to audio element
    else
        ivrAudio.duration = 0
    end

    -- search 'repeat' element
    eErrNo = xmlChildElementByName(fsm.xmlDoc,"repeat")
    if eErrNo==MSS_ERR_OK then
        ivrAudio.repeatCnt = ToInt(xmlGetElementData(fsm.xmlDoc))
        xmlParentElement(fsm.xmlDoc) -- back to audio element
    else
        ivrAudio.repeatCnt = 1
    end
    TraceInfo("\t\t\t id="..ivrAudio.id..",duration="..ivrAudio.duration..",repeat="..ivrAudio.repeatCnt)
    return ivrAudio
end

local function parseElementMonitorEvent()
    TraceInfo("\t\t parseElementMonitorEvent")
    local event={}
    
    -- search 'nextaction'
    event.nextaction = xmlGetElementAttr(fsm.xmlDoc, "nextaction")
    if event.nextaction==nil or string.len(event.nextaction)==0 then 
        TraceInfo("\t\t\t cannot find nextaction")
        return nil 
    end
    
    -- search 'detection'
    local detection = xmlGetElementAttr(fsm.xmlDoc, "detection")
    if detection==nil or string.len(detection)==0 then
        TraceInfo("\t\t\t cannot find detection")
        return nil
    end
    
    -- convert detection string to DP value
    if detection=="busy" then
        event.detection = DP_O_BUSY
    elseif detection=="no-answer" then
        event.detection = DP_O_NO_ANSWER
    elseif detection=="reject" then
        event.detection = DP_ROUTE_SELECT_FAILURE
    elseif detection=="disconnect" then
        event.detection = DP_O_DISCONNECT
    else
        TraceInfo("\t\t\t unknow detection?")
        return nil
    end

    TraceInfo("\t\t\t detection="..event.detection..", nextaction="..event.nextaction)
    return event
end


local function parseElementDigit()
    TraceInfo("\t\t parseElementDigit")
    local digit={}

    -- search 'nextaction'
    digit.nextaction = xmlGetElementAttr(fsm.xmlDoc, "nextaction")
    if digit.nextaction==nil then return nil end

    -- search 'value'
    digit.value = xmlGetElementAttr(fsm.xmlDoc, "value")
    if digit.value==nil or string.len(digit.value)==0 then
        digit.value = "default"
    end
    TraceInfo("\t\t\t value="..digit.value..", nextaction="..digit.nextaction)
    return digit
end


local function parseElementInput()
    TraceInfo("\t\t parseElementInput")
    local inputElem={}

    -- search 'from'
    local fromStr=xmlGetElementAttr(fsm.xmlDoc, "from")
    if fromStr~=nil and string.len(fromStr)>0 then
        inputElem.from = fromStr
        TraceInfo("\t\t\t input from "..fromStr)
    else
        TraceInfo("\t\t\t input from current PNC operation.")        
    end

    return inputElem
end

local function parseElementMonitorEvents()
    TraceInfo("\t\t parseElementMonitorEvents")
    local monitorEvents={}
    local eventCnt = xmlElementsByTagName(fsm.xmlDoc, "monitor-event")
    if eventCnt==0 then
        TraceInfo("\t\t\t no monitor-event.")
        return nil
    end    
        
    local events={}
    for index=1, eventCnt do
        local event = parseElementMonitorEvent()
        if event==nil then
            TraceInfo("\t\t\t fail to parse monitor-event")
            return nil
        end
        events[index]=event
        xmlNextElementByName(fsm.xmlDoc, "monitor-event")
    end
    xmlParentElement(fsm.xmlDoc) -- back to 'monitor-events' element
        
    monitorEvents.events = events    
    monitorEvents.eventCnt = eventCnt
    return monitorEvents    
end


local function parseElementValue()
    TraceInfo("\t\t parseElementValue")
    local value={}
    
    local eErrNo = xmlChildElementByName(fsm.xmlDoc, "callernbr")
    if eErrNo==MSS_ERR_OK then
        value.type = "callernbr"
        xmlParentElement(fsm.xmlDoc) -- back to 'value' element
        return value
    end

    eErrNo = xmlChildElementByName(fsm.xmlDoc, "callednbr")
    if eErrNo==MSS_ERR_OK then
        value.type = "callednbr"
        xmlParentElement(fsm.xmlDoc) -- back to 'value' element
        return value
    end

    eErrNo = xmlChildElementByName(fsm.xmlDoc, "input")
    if eErrNo==MSS_ERR_OK then        
        local inputElem = parseElementInput()
        value.type = "inputVal"
        value.content = inputElem
        xmlParentElement(fsm.xmlDoc) -- back to 'value' element
        return value
    end
    
    value.type = "stringVal"
    value.content = xmlGetElementData(fsm.xmlDoc)    
    return value
end


local function parseElementParameter()
    TraceInfo("\t\t parseElementParameter")
    local param={}

    -- search 'name'
    local eErrNo = xmlChildElementByName(fsm.xmlDoc,"name")
    if eErrNo~=MSS_ERR_OK then
        TraceInfo("\t\t\t cannot find 'name'.")
        return nil
    end
    param.name = xmlGetElementData(fsm.xmlDoc)
    xmlParentElement(fsm.xmlDoc) -- back to 'parameter' element

    -- search 'value'
    eErrNo = xmlChildElementByName(fsm.xmlDoc,"value")
    if eErrNo~=MSS_ERR_OK then
        TraceInfo("\t\t\t cannot find 'value'.")
        return nil
    end
    local value=parseElementValue()
    if value==nil then
        return nil
    end
    TraceInfo("\t\t\t value.type="..value.type)
    param.value = value
    xmlParentElement(fsm.xmlDoc) -- back to 'parameter' element

    return param
end


local function parseElementResult()
    TraceInfo("\t\t parseElementResult")
    local resultElem={}
    --search 'nextaction'
    local nextaction = xmlGetElementAttr(fsm.xmlDoc, "nextaction")
    if string.len(nextaction)==0 then
        TraceInfo("\t\t\t cannot find 'nextaction'.")
        return nil
    end
    resultElem.nextaction = nextaction

    -- search 'value'
    local value = xmlGetElementAttr(fsm.xmlDoc, "value")
    if string.len(value)==0 then
        value = "default"
    end
    resultElem.value = value
    TraceInfo("\t\t\t value="..value..",nextaction="..nextaction)
    return resultElem
end

local function parseActionPnC(currAction)
    TraceInfo("\t parseActionPnC")
    local eErrNo = xmlChildElementByName(fsm.xmlDoc,"playaudio")
    if eErrNo~=MSS_ERR_OK then
        TraceInfo("\t\t cannot find 'playaudio' element.") 
        return -1
    end
    local ivrAudio = parserElementAudio()
    if ivrAudio==nil then
        TraceInfo("\t\t cannot parser 'playaudio' element")
        return -2
    else
        currAction.playaudio = ivrAudio
        xmlParentElement(fsm.xmlDoc) -- back to PNC element
    end
    
    -- search 'minnumofdigits'
    eErrNo = xmlChildElementByName(fsm.xmlDoc, "minnumofdigits")
    if eErrNo==MSS_ERR_OK then
        currAction.minnumofdigits = ToInt(xmlGetElementData(fsm.xmlDoc))
        xmlParentElement(fsm.xmlDoc) -- back to PNC element
    else
        currAction.minnumofdigits = 1
    end
    TraceInfo("\t\t minnumofdigits="..currAction.minnumofdigits)

    -- search 'maxnumofdigits'
    eErrNo = xmlChildElementByName(fsm.xmlDoc, "maxnumofdigits")
    if eErrNo==MSS_ERR_OK then
        currAction.maxnumofdigits = ToInt(xmlGetElementData(fsm.xmlDoc))
        xmlParentElement(fsm.xmlDoc) -- back to PNC element
    else
        currAction.maxnumofdigits = 30
    end
    TraceInfo("\t\t maxnumofdigits="..currAction.maxnumofdigits)

    -- search 'firstdigittimeout' or "first-digit-timeout"
    eErrNo = xmlChildElementByName(fsm.xmlDoc, "firstdigittimeout")
    if eErrNo~=MSS_ERR_OK then
        eErrNo = xmlChildElementByName(fsm.xmlDoc, "first-digit-timeout")
    end
    if eErrNo==MSS_ERR_OK then
        currAction.firstdigittimeout = ToInt(xmlGetElementData(fsm.xmlDoc))
        xmlParentElement(fsm.xmlDoc) -- back to PNC element
    else
        currAction.firstdigittimeout = 30
    end
    TraceInfo("\t\t firstdigittimeout="..currAction.firstdigittimeout)

    -- search "intervaldigittimeout" or "interval-digit-timeout"
    eErrNo = xmlChildElementByName(fsm.xmlDoc, "intervaldigittimeout")
    if eErrNo~=MSS_ERR_OK then
        eErrNo = xmlChildElementByName(fsm.xmlDoc, "interval-digit-timeout")
    end
    if eErrNo==MSS_ERR_OK then
        currAction.intervaldigittimeout = ToInt(xmlGetElementData(fsm.xmlDoc))
        xmlParentElement(fsm.xmlDoc) -- back to PNC element
    else
        currAction.intervaldigittimeout = 10
    end
    TraceInfo("\t\t intervaldigittimeout="..currAction.intervaldigittimeout)

    -- search 'enddigit'
    eErrNo = xmlChildElementByName(fsm.xmlDoc, "enddigit")
    if eErrNo==MSS_ERR_OK then
        currAction.enddigit = xmlGetElementData(fsm.xmlDoc)
        xmlParentElement(fsm.xmlDoc) -- back to PNC element
    else
        currAction.enddigit="#"
    end
    TraceInfo("\t\t enddigit="..currAction.enddigit)

    -- search "canceldigit"
    eErrNo = xmlChildElementByName(fsm.xmlDoc, "canceldigit")
    if eErrNo==MSS_ERR_OK then
        currAction.canceldigit = xmlGetElementData(fsm.xmlDoc)
        xmlParentElement(fsm.xmlDoc)
    else
        currAction.canceldigit="*"
    end
    TraceInfo("\t\t canceldigit="..currAction.canceldigit)

    return 0
end


local function parseActionPA(currAction)
    TraceInfo("\t parseActionPA")
    local eErrNo = xmlChildElementByName(fsm.xmlDoc,"playaudio")
    if eErrNo~=MSS_ERR_OK then
        TraceInfo("\t\t cannot find 'playaudio' element.") 
        return -1
    end

    local ivrAudio = parserElementAudio()
    if ivrAudio==nil then
        TraceInfo("\t\t cannot parser 'playaudio' element")
        return -2
    end
    currAction.playaudio = ivrAudio
    xmlParentElement(fsm.xmlDoc) -- back to PA element
    
    return 0
end


local function parseActionInput(currAction)
    TraceInfo("\t parseActionInput")

    local digitCnt = xmlElementsByTagName(fsm.xmlDoc, "digit")
    if digitCnt==0 then
        TraceInfo("\t\t cannot find 'digit' elements")
        return -1
    end
    TraceInfo("\t\t find "..digitCnt.." digit element(s)")
    local digits={}
    for index=1, digitCnt do
        local digit = parseElementDigit()
        if digit==nil then
            TraceInfo("\t\t fail to parser digit element")
            return -2
        end
        digits[digit.value]=digit        
        xmlNextElementByName(fsm.xmlDoc, "digit")
    end
    currAction.digits = digits
    xmlParentElement(fsm.xmlDoc) -- back to "input" element

    -- search 'error-first-digit-timeout'
    local eErrNo = xmlChildElementByName(fsm.xmlDoc, "error-first-digit-timeout")
    if eErrNo==MSS_ERR_OK then
        currAction.errFirstDigitTimeout = xmlGetElementAttr(fsm.xmlDoc, "nextaction")
        TraceInfo("\t\t errFirstDigitTimeout="..currAction.errFirstDigitTimeout)
        xmlParentElement(fsm.xmlDoc)
    else
        currAction.errFirstDigitTimeout=nil
    end

    -- search 'error-improper-caller-response'
    eErrNo = xmlChildElementByName(fsm.xmlDoc, "error-improper-caller-response")
    if eErrNo==MSS_ERR_OK then
        currAction.errImproperCallerRespTimeout = xmlGetElementAttr(fsm.xmlDoc, "nextaction")
        TraceInfo("\t\t errImproperCallerRespTimeout="..currAction.errImproperCallerRespTimeout)
        xmlParentElement(fsm.xmlDoc)
    else
        currAction.errImproperCallerRespTimeout=nil
    end
    
    return 0
end


local function parseActionCallto(currAction)
    TraceInfo("\t parseActionCallto")

    local eErrNo = xmlChildElementByName(fsm.xmlDoc, "monitor-events")
    if eErrNo==MSS_ERR_OK then
        TraceInfo("\t\t find monitor-events element")
        currAction.monitorEvents = parseElementMonitorEvents()
        xmlParentElement(fsm.xmlDoc) -- back to 'callto' element
    end
    
    eErrNo = xmlChildElementByName(fsm.xmlDoc, "destination")
    if eErrNo~=MSS_ERR_OK then
        TraceInfo("\t\t cannot get 'destination' element")
        return -1
    end

    eErrNo = xmlChildElementByName(fsm.xmlDoc, "input")
    if eErrNo~=MSS_ERR_OK then
        currAction.number = xmlGetElementData(fsm.xmlDoc)
        TraceInfo("\t\t destination="..currAction.number)        
    else
        currAction.input = parseElementInput()
        xmlParentElement(fsm.xmlDoc) -- back to 'destination' element    
    end

    xmlParentElement(fsm.xmlDoc) -- back to 'callto' element
    return 0
end


local function parseActionRelCall(currAction)
    TraceInfo("\t parseActionRelCall")
    local eErrNo = xmlChildElementByName(fsm.xmlDoc, "cause")
    if eErrNo==MSS_ERR_OK then
        currAction.cause = ToInt(xmlGetElementData(fsm.xmlDoc))
        xmlParentElement(fsm.xmlDoc) -- back to 'relcall' element
    else
        currAction.cause = CAUSE_NORMAL_UNSPECIFIED        
    end
    return 0
end


local function parseActionGetURL(currAction)
    TraceInfo("\t parseActionGetURL")
    local eErrNo = xmlChildElementByName(fsm.xmlDoc, "url")
    if eErrNo~=MSS_ERR_OK then
        TraceInfo("\t\t cannot find 'url' element.")
        return -1
    end
    currAction.url = xmlGetElementData(fsm.xmlDoc)
    xmlParentElement(fsm.xmlDoc) -- back to 'geturl' element

    local paramCnt = xmlElementsByTagName(fsm.xmlDoc, "parameter")
    if paramCnt==0 then
        TraceInfo("\t\t no parameter.")
        return 0
    end
    currAction.paramCnt = paramCnt
    local params={}
    for index=1, paramCnt do
        local param = parseElementParameter()
        if param==nil then
            TraceInfo("\t\t fail to parse parameter")
            return -2
        end
        params[index]=param
        xmlNextElementByName(fsm.xmlDoc, "parameter")
    end
    currAction.params = params
    xmlParentElement(fsm.xmlDoc) -- back to "geturl" element

    return 0
end


local function parseActionGetURLResult(currAction)
    TraceInfo("\t parseActionGetURLResult")

    local resultCnt = xmlElementsByTagName(fsm.xmlDoc, "result")
    if resultCnt==0 then
        TraceInfo("\t\t cannot find 'result' elements")
        return -1
    end
    TraceInfo("\t\t find "..resultCnt.." result element(s)")
    local results={}
    for index=1, resultCnt do
        local result = parseElementResult()
        if result==nil then
            TraceInfo("\t\t fail to parse result element")
            return -2
        end
        results[result.value]=result
        xmlNextElementByName(fsm.xmlDoc, "result")
    end
    currAction.results =  results
    xmlParentElement(fsm.xmlDoc) -- back to "geturlResult" element
    return 0
end


local function parseAction(actionIndex)
    local actName=xmlGetElementAttr(fsm.xmlDoc, "name")
    local actMethod=xmlGetElementAttr(fsm.xmlDoc, "method")
    if actName==nil or actMethod==nil then return -1 end
    if string.len(actName)==0 or string.len(actMethod)==0 then return -2 end

    local currAction={}
    currAction.name = actName
    currAction.method = actMethod
    currAction.actionIndex = actionIndex
    local result = 0
    if actMethod==IVR_METHOD_PNC then
        result = parseActionPnC(currAction)
    
    elseif actMethod==IVR_METHOD_INPUT then
        result = parseActionInput(currAction)
    
    elseif actMethod==IVR_METHOD_CALLTO then
        result = parseActionCallto(currAction)

    elseif actMethod==IVR_METHOD_RELCALL then
        result = parseActionRelCall(currAction)

    elseif actMethod==IVR_METHOD_PA then
        result = parseActionPA(currAction)
    
    elseif actMethod==IVR_METHOD_GETURL then
        result = parseActionGetURL(currAction)
    
    elseif actMethod==IVR_METHOD_GETURL_RESULT then
        result = parseActionGetURLResult(currAction)

    else
        TraceInfo("\t\t unknown method: "..actMethod)
        return -3
    end

    if result<0 then return -4 end

    fsm.actionDict[actName]=currAction
    fsm.actionNameList[actionIndex]=actName
    return 0
end


local function parseActions()
    TraceInfo("\t parseActions")
    for actionIndex=1,fsm.actionCnt do
        local result = parseAction(actionIndex)
        if result<0 then return result end

        xmlNextElementByName(fsm.xmlDoc, "action")
    end
    return 0
end


local function parseXMLFile()
    TraceInfo("\t parseXMLFile")
    local custIvrPath = ssfGetCustIvrXmlDir()
    local ivrxmlFullName = custIvrPath..fsm.ivrxmlFileName    
    fsm.xmlDoc = xmlOpen(ivrxmlFullName)
    if fsm.xmlDoc==nil then
        local sysIvrPath = ssfGetSysIvrXmlDir()
        ivrxmlFullName = sysIvrPath..fsm.ivrxmlFileName
        fsm.xmlDoc = xmlOpen(ivrxmlFullName)
        if fsm.xmlDoc==nil then
            TraceInfo("\t\t fail to parser XML file: "..fsm.ivrxmlFileName) 
            return -1 
        end
    end
    TraceInfo("\t\t prepare to parser "..ivrxmlFullName)
    
    fsm.actionCnt = xmlElementsByTagName(fsm.xmlDoc, "action")
    if fsm.actionCnt==0 then
        TraceInfo("\t\t cannot find action.")
        xmlClose(fsm.xmlDoc)
        return -2
    end
    TraceInfo("\t\t find "..fsm.actionCnt.." action(s)")

    local result = parseActions()
    if result<0 then
        TraceInfo("\t\t fail to parser actions. result="..result)
        xmlClose(fsm.xmlDoc)
        return -3 
    end
    
    xmlClose(fsm.xmlDoc)
    return 0
end


--====================================
--
-- Actions functions
--
--====================================
function prepareNextAction()
    TraceInfo("\t prepareNextAction")

    fsm.previousActionIndex = fsm.currActionIndex
    fsm.previousActionMethod = fsm.currActionMethod
    fsm.currActionIndex = fsm.currActionIndex + 1
end


local function convertPnCArg(currAction, pncArg)
    local playAudio = currAction.playaudio
    pncArg.infoSend_annID_annID = playAudio.id
    pncArg.infoSend_duration = playAudio.duration
    pncArg.infoSend_numOfRepeat = playAudio.repeatCnt

    pncArg.minNumOfDigits = currAction.minnumofdigits
    pncArg.maxNumOfDigits = currAction.maxnumofdigits
    pncArg.firstDigitTimeout = currAction.firstdigittimeout
    pncArg.intervalDigitTimeout = currAction.intervaldigittimeout
    
    pncArg.endDigit = currAction.enddigit
    pncArg.cancelDigit = currAction.canceldigit
    return MSS_ERR_OK
end


local function convertPAArg(currAction, paArg)
    local playAudio = currAction.playaudio
    paArg.infoSend_annID_annID = playAudio.id
    paArg.infoSend_duration = playAudio.duration
    paArg.infoSend_numOfRepeat = playAudio.repeatCnt
    return MSS_ERR_OK
end


local function procActionPNC(currAction)
    TraceInfo("\t procActionPNC")
    local pncArg = CreatePNC()
    local eErrNo = convertPnCArg(currAction, pncArg)
    if eErrNo~=MSS_ERR_OK then
        TraceInfo("\t\t fail to convert PNC.")
        return MSS_ERR_INVALID_OPR
    end
    
    prepareIVR()    
    SendPNCWithArg(CALLER_LEGID, pncArg)
    prepareNextAction()

    StartTimer(IVR_TIMER_INTERACTION)
    EnterState(EIVR_INTERACTION)
    return MSS_ERR_OK
end


local function procActionInput(currAction)
    TraceInfo("\t procActionInput")

    local nextActionName=""
    if fsm.srfError==ESRF_ERR_OK then
        -- in 'parseActionInput', action use digit value as key
        local digit = currAction.digits[fsm.pncResult]
        if digit==nil then
            digit = currAction.digits["default"]
        end
        if digit~=nil then
            nextActionName = digit.nextaction
        end

    elseif fsm.srfError==ESRF_ERR_FIRST_DIGIT_TIMEOUT then
        TraceInfo("\t\t to process first digit timeout")
        if currAction.errFirstDigitTimeout~=nil then
            nextActionName = currAction.errFirstDigitTimeout
        else
            TraceInfo("\t\t cannot find action to process it")
            return MSS_ERR_INVALID_OPR
        end

    elseif fsm.srfError==ESRF_ERR_IMPROPER_CALLER_RESPONSE then
        TraceInfo("\t\t to process improper caller response")
        if currAction.errImproperCallerRespTimeout~=nil then
            nextActionName = currAction.errImproperCallerRespTimeout
        else
            TraceInfo("\t\t cannot find action to process it")
            return MSS_ERR_INVALID_OPR
        end

    else
        TraceInfo("\t\t don't know how to proces this action.")
        return MSS_ERR_INVALID_OPR
    end

    if string.len(nextActionName)==0 then
        TraceInfo("\t\t cannot find next action name. Just prepare next action directly.")
        prepareNextAction()
    else
        TraceInfo("\t\t prepare next action: "..nextActionName)
        fsm.previousActionMethod = IVR_METHOD_INPUT
        local nextAction = fsm.actionDict[nextActionName]
        if nextAction~=nil then
            fsm.currActionIndex = nextAction.actionIndex
        else
            TraceInfo("\t\t cannot find next action")
            return MSS_ERR_INVALID_OBJ
        end
    end

    return procCurrAction()
end


local function procActionCallto(currAction)
    TraceInfo("\t procActionCallto")

    local destNbr=""
    if currAction.number~=nil then
        destNbr = currAction.number
    elseif currAction.input~=nil then
        destNbr = getInputResult(currAction.input)
    else
        return MSS_ERR_INVALID_OBJ
    end

    if string.len(destNbr)==0 then
        TraceInfo("\t\t cannot get destination number")
        return MSS_ERR_OBJ_NOT_FOUND
    end
    TraceInfo("\t\t call to "..destNbr)
    
    SendDisarmAll() -- Clear all previous EDPs
    monitorCall() -- monitor o-answer to stop ring-back tone
    
    if currAction.monitorEvents~=nil then
        local monitorEvents = currAction.monitorEvents
        for index=1, monitorEvents.eventCnt do
            local event = monitorEvents.events[index]
            SendRRBE(CALLED_LEGID, event.detection, DP_MON_INTERUPT)
        end
    end
    
    -- prepare IVR to play ring-back tone
    prepareIVR()    
    SendRingBackToneToCaller()    
    SendConnect(destNbr)

    StartTimer(IVR_TIMER_WAIT_ANSWER)
    EnterState(EIVR_WAIT_ANSWER)
    return MSS_ERR_OK
end


local function procActionPA(currAction)
    TraceInfo("\t procActionPA")
    local paArg=CreatePA()
    convertPAArg(currAction, paArg)
    
    prepareIVR()
    SendPA(CALLER_LEGID, paArg)

    StartTimer(IVR_TIMER_INTERACTION)
    EnterState(EIVR_INTERACTION)
    return MSS_ERR_OK
end


local function procActionRelCall(currAction)
    TraceInfo("\t procActionRelCall")

    SendReleaseCall(currAction.cause)
    EndService()
    return MSS_ERR_OK 
end


local function procActionGetURL(currAction)
    TraceInfo("\t procActionGetURL")
    local fullURL=getFullURL(currAction)
    if string.len(fullURL)==0 then
        TraceInfo("\t\t blank URL??")
        return MSS_ERR_INVALID_OBJ
    end

    fsm.geturlResult = ssfGetURL(fullURL)
    TraceInfo("\t\t URL result: "..fsm.geturlResult)

    prepareNextAction()    
    return procCurrAction()
end


local function procActionGetURLResult(currAction)
    TraceInfo("\t procActionGetURLResult")

    -- save original geturl result
    currAction.result = fsm.geturlResult

    -- get default next action    
    local results = currAction.results
    local result = results[currAction.result]
    if result==nil then
        result = results["default"]
        if result==nil then
            TraceInfo("\t\t cannot find the result")
            return MSS_ERR_OBJ_NOT_FOUND
        end
    end
    local nextActionName=result.nextaction
    if string.len(nextActionName)==0 then
        TraceInfo("\t\t  cannot find next action name, go to next action directly.")
        prepareNextAction()
    else
        TraceInfo("\t\t prepare next action: "..nextActionName)
        local nextAction = fsm.actionDict[nextActionName]
        if nextAction==nil then
            TraceInfo("\t\t cannot find next action")
            return MSS_ERR_OBJ_NOT_FOUND
        else
            fsm.currActionIndex = nextAction.actionIndex
        end
    end
    return procCurrAction()
end


function procCurrAction()
    TraceInfo("\t procCurrAction")

    local currActionName = fsm.actionNameList[fsm.currActionIndex]    
    local currAction = fsm.actionDict[currActionName]
    if currAction==nil then
        TraceInfo("\t\t cannot find action.")
        return MSS_ERR_OBJ_NOT_FOUND
    end

    local currMethod = currAction.method
    fsm.currActionMethod = currMethod
    if currMethod==IVR_METHOD_PNC then
        return procActionPNC(currAction)

    elseif currMethod==IVR_METHOD_INPUT then
        return procActionInput(currAction)

    elseif currMethod==IVR_METHOD_CALLTO then
        return procActionCallto(currAction)
    
    elseif currMethod==IVR_METHOD_PA then
        return procActionPA(currAction)
    
    elseif currMethod==IVR_METHOD_RELCALL then
        return procActionRelCall(currAction)
    
    elseif currMethod==IVR_METHOD_GETURL then
        return procActionGetURL(currAction)
    
    elseif currMethod==IVR_METHOD_GETURL_RESULT then
        return procActionGetURLResult(currAction)

    else
        TraceInfo("\t\t unknown method: "..currMethod)
        return MSS_ERR_UNKNOWN
    end
    return MSS_ERR_OK
end


--====================================
--
-- FSM state functions
--
--====================================
local function interactionPnCResult()
    TraceInfo("\t interactionPnCResult")
    TraceInfo("\t\t pnc result is "..fsm.pncResult)

    local currActionName = fsm.actionNameList[fsm.currActionIndex]    
    local currAction = fsm.actionDict[currActionName]
    local currMethod = currAction.method
    if currMethod~=IVR_METHOD_INPUT then
        TraceInfo("\t\t current action is not 'input'")
        return MSS_ERR_INVALID_OPR
    end
    -- save PnC result to current 'input' action
    fsm.actionDict[currActionName].result = fsm.pncResult

    local result = procCurrAction()
    if result~=MSS_ERR_OK then
        return MSS_ERR_FAIL
    else
        return MSS_ERR_OK
    end
end


local function interactionSRFError()
    TraceInfo("\t interactionSRFError")

    local result = procCurrAction()
    if result~=MSS_ERR_OK then
        return MSS_ERR_FAIL
    end
    return MSS_ERR_OK
end


local function interactionSRR()
    TraceInfo("\t interactionSRR")
    prepareNextAction()

    local result = procCurrAction()
    if result~=MSS_ERR_OK then
        return MSS_ERR_FAIL
    end
    return MSS_ERR_OK
end


local function onIdleProc(currMsg)
    TraceInfo("IVR-XML: onIdleProc")
    initData()  
    local result = getCustIVRXML()
    if result<0 then
        exception()
        return
    end

    result = parseXMLFile()
    if result<0 then
        TraceInfo("\t fail to parser XML file.")
        exception()
        return
    end
    
    result = procCurrAction()
    if result~=MSS_ERR_OK then
        exception()
        return
    end
end


local function onInteractionProc(currMsg)
    TraceInfo("onInteractionProc")
    local eErrNo = MSS_ERR_OK
    if currMsg==ESCP_EVT_PNC_RESULT then
        fsm.srfError = ESRF_ERR_OK
        fsm.pncResult = currMessage.digits
        eErrNo = interactionPnCResult()
    
    elseif currMsg==ESCP_EVT_SRF_ERR then        
        fsm.srfError = currMessage.srfError
        fsm.pncResult=""
        eErrNo = interactionSRFError()

    elseif currMsg==ESCP_EVT_SRR then
        eErrNo = interactionSRR()

    else
        exception()
        return
    end

    if eErrNo~=MSS_ERR_OK then
        exception()
    end
end

-- ------------------------------------------------------------------------------------------------------
-- 20220905 If called side is busy or no-answer, it is possible that ivr to play 
--          audio to caller party, at this time, the ivr still plays ring-back tone,
--          so it is necessary to stop ivr firstly.
-- 20220906 If receives EDP_PROGRESS, should not disarm all EDPs, just stop IVR and continue SSF call.
-- ------------------------------------------------------------------------------------------------------
local function onWaitAnswerProc(currMsg)
    TraceInfo("onWaitAnswerProc")
    if currMsg~=ESCP_EVT_ERB then
        TraceInfo("\t Not event-report, stop IVR-XML.")
        exception()
        return
    end
    TraceInfo("\t current EDP is "..currMessage.edpEvent)

    -- process some EDPs
    if currMessage.edpEvent==DP_O_CALL_PROGRESS then
        TraceInfo("\t recv call progress from called side")
        if currMessage.sdpInd==MSS_YES then
            stopIVR()
        end
        SendContinue()
        return
    end

    if currMessage.edpEvent==DP_O_ANSWER then 
        TraceInfo("\t Called side answers the call.")
        fsm.connFailCnt = 0
        stopIVR()
        SendDisarmAll()
        SendContinue()
        EndService()
        return
    end    
    
    local currActionName = fsm.actionNameList[fsm.currActionIndex]    
    local currAction = fsm.actionDict[currActionName]
    if currAction==nil or fsm.currActionMethod~=IVR_METHOD_CALLTO then
        TraceInfo("\t current action is not callto, something wrong.")
        exception()
        return
    end
    
    local monitorEvents = currAction.monitorEvents    
    if monitorEvents==nil then
        TraceInfo("\t no EDP armed, release the call.")
        exception()
        return
    end
    
    local eventCnt = monitorEvents.eventCnt
    local events = monitorEvents.events
    if eventCnt==nil or events==nil then
        TraceInfo("\t no event, release the call.")
        exception()
        return
    end
    
    for index=1, eventCnt do
        local event = events[index]
        if event~=nil and event.detection==currMessage.edpEvent then
            local nextActionName = event.nextaction
            local nextAction = fsm.actionDict[nextActionName]
            if nextAction~=nil then
                fsm.connFailCnt = fsm.connFailCnt + 1
                if fsm.connFailCnt>MAX_FAIL_CONN_CNT then
                    TraceInfo("\t too many failed connection, stop IVR_XML.")
                    exception()
                    return
                end

                -- for o-busy or o-no-answer, free ring-back tone firstly
                if currMessage.edpEvent==DP_O_BUSY or currMessage.edpEvent==DP_O_NO_ANSWER then
                    stopIVR()
                end
                
                fsm.currActionIndex = nextAction.actionIndex
                procCurrAction()
                return
            else
                TraceInfo("\t cannot find next action")
                exception()
                return
            end
        end
    end
    
    TraceInfo("\t doesn't care current event.")
    exception()
end

--====================================
--
-- Service main function
--
--====================================
function DestroyProc()
    TraceInfo("DestroyProc")    
end


function MainProc(currMsg)      
    StopTimer()
    if fsm.state==EIVR_IDLE then
        onIdleProc(currMsg)

    elseif fsm.state==EIVR_INTERACTION then
        onInteractionProc(currMsg)

    elseif fsm.state==EIVR_WAIT_ANSWER then
        onWaitAnswerProc(currMsg)

    else
        exception()
    end    
end
